/* Filename: Hash_tree.c*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "MD5.h"
#include "Hash_tree.h"


typedef struct tree_node *tree_nodeptr;

/*domi typou komvou tou dentrou*/
typedef struct tree_node
{
    unsigned int digest[4];
    int id;
    int branches_num;
    tree_nodeptr *branches;
    tree_nodeptr parent;    
}tree_node;

/*domi typou kefalis dentrou*/
typedef struct tree_head
{
    int depth;   
    int leafs;
    int fanout;
    tree_nodeptr root;
    tree_nodeptr current;    
}tree_head;


/*synartiseis pou xrisimopoiountai topika*/

/*eisagei to digest sto dentro*/
int Hash_tree_insert(Hash_tree tree, unsigned int digest[4], MD5_CTX *context);
/*elengxei an o node exei digest*/
int has_digest(tree_nodeptr node);
/*anadromiki synartisi pou katastrefei tous nodes tou dentrou*/
void nodes_destroy(tree_nodeptr node);
/*ylopoiei ton anadromiko algorithmo sygrisis gia tous dyo komvous*/
void comparison(tree_nodeptr node1, tree_nodeptr node2);


int Hash_tree_create(Hash_tree *tree, int tree_fanout)
{
    /*desmeusi tis kefalis tou dentrou*/
    *tree = malloc(sizeof(tree_head));   
    if(tree == NULL)
    {
        printf("Cannot allocate memory\n");
        return 1;
    }
    
    /*arxikopoiisi ton metavliton tis*/
    (*tree)->root = NULL;    
    (*tree)->current = NULL;
    (*tree)->leafs = 0;
    (*tree)->depth = 0;
    (*tree)->fanout = tree_fanout;
    
    return 0;
}


void Hash_tree_destroy(Hash_tree *tree)
{  
    printf("\nDestroying tree\n"); 
    
    /*an to dentro den einai keno*/
    if((*tree)->root != NULL)
    {
        /*kalei tin synartisi katastrofis ton komvon*/
        nodes_destroy((*tree)->root);
    }
    
    /*eleutheronei tin kefali*/
    free(*tree);
    *tree = NULL;
}

void nodes_destroy(tree_nodeptr node)
{
    int i;
    
    /*kaleitai anadromika gia kathe paidi tou trexonta komvou*/
    for(i = 0; i< node->branches_num; i++)    
        nodes_destroy(node->branches[i]);  
    
    /*an den einai fyllo o trexontas komvos*/
    if(node->branches != NULL)
        /*eleutheronei ton pinaka me ta paidia tou*/
        free(node->branches);        

    /*eleutheronei ton trexonta komvo*/
    free(node);    
}


int Hash_tree_generate(char* file, Hash_tree tree, int page_size)
{
    printf("\nGenerating tree\n"); 
    
    FILE *fp;
    size_t read;
    char* buffer = malloc(page_size*sizeof(char));
    if(buffer == NULL)
    {
        printf("Cannot allocate memory\n");
        return 1;
    }
    
    tree_nodeptr tempnode;
    MD5_CTX context;
    unsigned int file_digest[4];  
    int i;

    fp = fopen(file, "rb");
    if (fp == NULL)
    {
        perror("Could not open file");
        free(buffer);
        return 1;
    }
    
    printf("File %s:\n\n", file);

    i = 1;
    do
    { 
        read = fread(buffer, 1, page_size, fp);        
        if (read > 0)
        {
            if (read < page_size)
            {
                memset(buffer + read, 0, page_size - read);
            }            
            printf("Buffer: %s\n", buffer);
            
            /*ypologismos tis sinopsis (digest)*/
            MD5Init(&context);
            MD5Update(&context, buffer, strlen(buffer));
            MD5Final(file_digest, &context);
            
            /*Ektiposi tis sinopsis*/            
            printf("MD5=");
            MD5Print( file_digest );
            printf("\n");            
            
            /*eisagogi tis synopsis sto dentro*/
            if(Hash_tree_insert(tree, file_digest, &context))
            {
                free(buffer);
                return 1;
            }                       
            
        }
        else
        {
            if(i == 1)
            {
                printf("Empty file!\n");
                return 0;
            }
        }
        i++;
    }while (read == page_size);

    /*an i riza den exei digest epeidi o teleutaios komvos den exei ola ta paidia*/
    if(!has_digest(tree->root))
    { 
        /*paei ston proto xoris digest*/
        tempnode = tree->current;
        while(has_digest(tempnode))
        {
            tempnode = tempnode->parent;
        }        
        
        while(1)
        {
            /*ypologismos tis synopsis apo ola ta paidia tou*/
            MD5Init( &context );
            printf("Finalizing Concatenation of: ");
            for(i = 0; i < tempnode->branches_num; i++)
            {
                MD5Update( &context , (char *)tempnode->branches[i] , 4 * sizeof( int ) );
                printf(" %d ",tempnode->branches[i]->id);
            }
            printf("\n");
            MD5Final( tempnode->digest , &context );

            /*Ektiposi tis sinopsis*/            
            printf("MD5=");
            MD5Print( tempnode->digest );
            printf("\n");
            
            /*otan ftasei sti riza stamataei*/
            if(tempnode->parent == NULL)
                break;        
            
            /*pairnei ton parapano komvo*/
            tempnode = tempnode->parent;
        }                 
    }    
    
    if (ferror(fp))
        perror("An error occured during reading");
    
    fclose(fp); 
    
    free(buffer);  
    buffer = NULL;
    
    return 0;
}


int Hash_tree_insert(Hash_tree tree, unsigned int digest[4], MD5_CTX *context)
{
    tree_nodeptr root, tempnode1, tempnode2, leaf;
    int i, j;
          
    /*an einai i proti eisodos sto dentro*/
    if(tree->depth == 0)
    {
        /*dimiourgei kai arxikopoiei tin riza*/
        root = malloc(sizeof(tree_node));
        if(root == NULL)
        {
            printf("Cannot allocate memory\n");
            return 1;
        }       
        
        root->branches = malloc(tree->fanout*sizeof(tree_nodeptr));
        if(root->branches == NULL)
        {
            printf("Cannot allocate memory\n");
            return 1;
        }           
        root->branches_num = 0;
        for(i = 0; i < 4; i++)
            root->digest[i] = 0;
        root->id = -1;
        root->parent = NULL;
        
        /*thetei analogos kai tis metavlites tis kefalis*/
        tree->root = root;  
        tree->current = root;
        tree->depth++;
    }
    
    /*an i riza exei digest tote ginetai diaspasi tis rizas*/
    if(has_digest(tree->root))
    {     
        /*dimiourgei kai arxikopoiei tin nea riza*/
        root = malloc(sizeof(tree_node));
        if(root == NULL)
        {
            printf("Cannot allocate memory\n");
            return 1;
        } 
        
        root->branches = malloc(tree->fanout*sizeof(tree_nodeptr));
        if(root->branches == NULL)
        {
            printf("Cannot allocate memory\n");
            return 1;
        }           
        root->branches_num = 1;
        for(i = 0; i < 4; i++)
            root->digest[i] = 0;
        root->id = -(tree->depth)-1;
        root->parent = NULL;        
        root->branches[0] = tree->root;        
        
        /*dimiourgei kai arxikopoiei tosous komvous osous to vathos tou dentrou*/
        tempnode1 = root;
        for(i = 0; i < tree->depth; i++)
        {
            tempnode2 = malloc(sizeof(tree_node));
            if(tempnode2 == NULL)
            {
                printf("Cannot allocate memory\n");
                return 1;
            } 

            tempnode2->branches = malloc(tree->fanout*sizeof(tree_nodeptr));
            if(tempnode2->branches == NULL)
            {
                printf("Cannot allocate memory\n");
                return 1;
            }           
            tempnode2->branches_num = 0;
            for(j = 0; j < 4; j++)
                tempnode2->digest[j] = 0;
            tempnode2->parent = tempnode1; 
            tempnode2->id = (tempnode2->parent->id) + 1;                   
            
            tempnode1->branches[tempnode1->branches_num] = tempnode2;
            tempnode1->branches_num++;
            
            tempnode1 = tempnode2;            
        }
        
        /*thetei to parent tis palias rizas na einai i nea riza*/
        tree->root->parent = root;    
        /*thetei analogos kai tis metavlites tis kefalis*/
        tree->root = root;  
        tree->current = tempnode1;
        tree->depth++;             
    }
    
    /*anevainei sto dentro oso oi komvoi exoun digest*/
    tempnode1 = tree->current;
    while(has_digest(tempnode1))
    {
        tempnode1 = tempnode1->parent;
    }
    
    /*an den einai o trexontas komvos teleutaiou epipedou*/
    while(tempnode1->id != -1)
    {      
        /*diiourgei kai arxikopoei kovous mexri na ftasei sto teleutaio epipedo*/
        tempnode2 = malloc(sizeof(tree_node));
        if(tempnode2 == NULL)
        {
            printf("Cannot allocate memory\n");
            return 1;
        } 

        tempnode2->branches = malloc(tree->fanout*sizeof(tree_nodeptr));
        if(tempnode2->branches == NULL)
        {
            printf("Cannot allocate memory\n");
            return 1;
        }           
        tempnode2->branches_num = 0;
        for(j = 0; j < 4; j++)
            tempnode2->digest[j] = 0;
        tempnode2->parent = tempnode1; 
        tempnode2->id = (tempnode2->parent->id) + 1;                   

        tempnode1->branches[tempnode1->branches_num] = tempnode2;
        tempnode1->branches_num++;

        tempnode1 = tempnode2;           
        
        if(tempnode1->id == -1)
        {       
            /*thetei os trexonta komvo ton teleutaio pou dimiourgithike*/
            tree->current = tempnode1;
        }
    }    
    
    /*dimiourgei kai arxikopoiei to fyllo pou tha mpei to digest*/
    leaf = malloc(sizeof(tree_node));
    if(leaf == NULL)
    {
        printf("Cannot allocate memory\n");
        return 1;
    }

    leaf->branches = NULL;          
    leaf->branches_num = 0;
    for(i = 0; i < 4; i++)
        leaf->digest[i] = digest[i];
    leaf->id = ++tree->leafs;
    leaf->parent = tree->current;

    /*thetei analogos tis metavlite tou komvou pou prostethike to fyllo*/
    tree->current->branches[leaf->parent->branches_num] = leaf;     
    tree->current->branches_num++;  

    /*oso oi komvoi einai gematoi ypologizei to digest concatenation tous*/
    tempnode1 = tree->current;
    while(tempnode1->branches_num == tree->fanout)
    {        
        MD5Init( context );
        printf("Concatenation of: ");
        for(i = 0; i < tree->fanout; i++)
        {
            MD5Update( context , (char *)tempnode1->branches[i] , 4 * sizeof( int ) );
            printf(" %d ",tempnode1->branches[i]->id);
        }
        printf("\n");
        MD5Final( tempnode1->digest , context );
        
        //Ektiposi tis sinopsis            
        printf("MD5=");
        MD5Print( tempnode1->digest );
        printf("\n");    
        
        /*an ftasei sti riza stamataei*/
        if(tempnode1->parent == NULL)
            break;        
        
        tempnode1 = tempnode1->parent;
    }
    
    return 0;
}


void Hash_tree_compare(Hash_tree tree1, Hash_tree tree2)
{
    printf("\nComparing trees\n");
    int i;
    
    /*an to dentro einai keno*/
    if(tree1->root == NULL)
    {
        printf("The files are empty so there aren't any differences!\n");
        return;
    }    
    
    /*sygrinei tis rizes ton dentron*/
    for(i = 0; i < 4; i++)   
        if(tree1->root->digest[i] != tree2->root->digest[i])
            break;

    /*an einai idies*/
    if(i == 4)
        printf("No difference\n");
    /*allios kalei tin comparison me tis rizes ton dyo dentron*/
    else    
        comparison(tree1->root, tree2->root);
}

void comparison(tree_nodeptr node1, tree_nodeptr node2)
{        
    int i;
    
    /*an einai se fyllo*/
    if(node1->branches == NULL)
    {     
        /*sygrinei ta digests ton fyllon*/
        for(i = 0; i < 4; i++)   
            if(node1->digest[i] != node2->digest[i])
                break;

        /*an den einai idia ektyponei ton arihmo selidas me tin diafora*/
        if(i < 4)
            printf("Difference in page %d\n", node1->id);            
    }
    else
    {
        /*sygrinei ta digests ton komvon*/
        for(i = 0; i < 4; i++)   
            if(node1->digest[i] != node2->digest[i])
                break;

        /*an einai idia stamataei*/
        if(i == 4)
            return;
        /*allios kalei ti comparison gia ta paidia tou komvou autou*/
        else
        {
            for(i = 0; i< node1->branches_num; i++)
                comparison(node1->branches[i], node2->branches[i]);            
        }
    }       
}

int has_digest(tree_nodeptr node)
{
    int i;
    
    for(i = 0; i < 4; i++)   
        if(node->digest[i] != 0)
            break;
    
    if(i == 4)
        return 0;
    else
        return 1;            
}
